#include <msp/gl/renderer.h>
#include "libr2c2/terraintype.h"
#include "layout.h"
#include "terrain.h"

using namespace Msp;

namespace R2C2 {

unsigned Terrain3D::surface_indices[] =
{
	0, 1, 3, 0, 3, 2,
	1, 3, 2, 1, 2, 0
};

unsigned Terrain3D::wall_indices[] =
{
	1, 4, 5, 1, 5, 3,
	3, 7, 6, 3, 6, 2
};

Terrain3D::Terrain3D(Layout3D &l, Terrain &t):
	Object3D(l, t),
	terrain(t),
	mesh((GL::VERTEX3, GL::NORMAL3)),
	tech(layout.get_catalogue().get<GL::Technique>("terrain.tech"))
{
	terrain.signal_size_changed.connect(sigc::mem_fun(this, &Terrain3D::size_changed));
	terrain.signal_tile_changed.connect(sigc::bind<2>(sigc::mem_fun(this, &Terrain3D::tile_changed), true));

	create_mesh();

	layout.get_scene().add(*this);
}

Terrain3D::~Terrain3D()
{
	layout.get_scene().remove(*this);
}

Vector Terrain3D::get_node() const
{
	float ts = terrain.get_type().get_tile_size();
	return terrain.get_position()+Vector(terrain.get_width()*ts/2, terrain.get_height()*ts/2, 0);
}

void Terrain3D::create_mesh()
{
	mesh.clear();

	GL::MeshBuilder bld(mesh);
	unsigned width = terrain.get_width();
	unsigned height = terrain.get_height();
	bld.begin(GL::TRIANGLES);
	for(unsigned y=0; y<height; ++y)
		for(unsigned x=0; x<width; ++x)
		{
			const Terrain::Tile &tile = terrain.get_tile(x, y);
			Vector nodes[8];
			gather_nodes(x, y, nodes);

			const unsigned *indices = surface_indices+tile.secondary_axis*6;
			build_triangle(bld, nodes, indices);
			build_triangle(bld, nodes, indices+3);

			if(x+1<width)
			{
				build_triangle(bld, nodes, wall_indices);
				build_triangle(bld, nodes, wall_indices+3);
			}

			if(y+1<height)
			{
				build_triangle(bld, nodes, wall_indices+6);
				build_triangle(bld, nodes, wall_indices+9);
			}
		}
	bld.end();
}

void Terrain3D::gather_nodes(unsigned x, unsigned y, Vector *nodes)
{
	const Terrain::Tile &tile = terrain.get_tile(x, y);
	float ts = terrain.get_type().get_tile_size();

	for(unsigned i=0; i<4; ++i)
		nodes[i] = Vector((x+i%2)*ts, (y+i/2)*ts, tile.nodes[i].elevation);

	if(x+1<terrain.get_width())
	{
		const Terrain::Tile &neighbor = terrain.get_tile(x+1, y);
		nodes[4] = Vector((x+1)*ts, y*ts, neighbor.nodes[0].elevation);
		nodes[5] = Vector((x+1)*ts, (y+1)*ts, neighbor.nodes[2].elevation);
	}

	if(y+1<terrain.get_height())
	{
		const Terrain::Tile &neighbor = terrain.get_tile(x, y+1);
		nodes[6] = Vector(x*ts, (y+1)*ts, neighbor.nodes[0].elevation);
		nodes[7] = Vector((x+1)*ts, (y+1)*ts, neighbor.nodes[1].elevation);
	}
}

void Terrain3D::build_triangle(GL::PrimitiveBuilder &bld, const Vector *nodes, const unsigned *indices)
{
	bld.normal(normalize(cross(nodes[indices[1]]-nodes[indices[0]], nodes[indices[2]]-nodes[indices[0]])));
	for(unsigned i=0; i<3; ++i)
		bld.vertex(nodes[indices[i]]);
}

void Terrain3D::update_triangle(unsigned offset, const Vector *nodes, const unsigned *indices)
{
	const GL::VertexFormat &format = mesh.get_vertices().get_format();
	unsigned position_offset = format.offset(GL::VERTEX3);
	unsigned normal_offset = format.offset(GL::NORMAL3);

	Vector normal = normalize(cross(nodes[indices[1]]-nodes[indices[0]], nodes[indices[2]]-nodes[indices[0]]));
	for(unsigned i=0; i<3; ++i)
	{
		float *vertex = mesh.modify_vertex(offset+i);
		for(unsigned j=0; j<3; ++j)
		{
			vertex[position_offset+j] = nodes[indices[i]][j];
			vertex[normal_offset+j] = normal[j];
		}
	}
}

void Terrain3D::size_changed(unsigned, unsigned)
{
	create_mesh();
}

void Terrain3D::tile_changed(unsigned x, unsigned y, bool propagate)
{
	Vector nodes[8];
	gather_nodes(x, y, nodes);

	unsigned base = y*(terrain.get_width()*18-6);
	if(y+1==terrain.get_height())
		base += x*12;
	else
		base += x*18;

	const Terrain::Tile &tile = terrain.get_tile(x, y);
	const unsigned *indices = surface_indices+tile.secondary_axis*6;
	update_triangle(base, nodes, indices);
	update_triangle(base+3, nodes, indices+3);

	if(x+1<terrain.get_width())
	{
		base += 6;
		update_triangle(base, nodes, wall_indices);
		update_triangle(base+3, nodes, wall_indices+3);
	}

	if(y+1<terrain.get_height())
	{
		base += 6;
		update_triangle(base, nodes, wall_indices+6);
		update_triangle(base+3, nodes, wall_indices+9);
	}

	if(propagate)
	{
		if(x>0)
			tile_changed(x-1, y, false);
		if(y>0)
			tile_changed(x, y-1, false);
	}
}

void Terrain3D::render(GL::Renderer &renderer, const GL::Tag &tag) const
{
	if(!tech.has_pass(tag))
		return;

	GL::Renderer::Push push(renderer);

	renderer.matrix_stack() *= matrix;

	const GL::RenderPass &pass = tech.get_pass(tag);
	renderer.set_material(pass.get_material());
	renderer.set_texturing(pass.get_texturing());
	renderer.set_shader_program(pass.get_shader_program(), pass.get_shader_data());

	mesh.draw(renderer);
}

} // namespace R2C2
