#include <msp/strings/format.h>
#include "libr2c2/articlenumber.h"
#include "designer.h"
#include "extendtool.h"

using namespace std;
using namespace Msp;
using namespace R2C2;

ExtendTool::ExtendTool(Designer &d, Input::Keyboard &k, Input::Mouse &m, const set<Object *> &objects):
	Tool(d, k, m),
	max_preference(0)
{
	for(set<Object *>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		if(Track *t = dynamic_cast<Track *>(*i))
		{
			unsigned nls = t->get_n_link_slots();
			for(unsigned j=0; j<nls; ++j)
				if(!t->get_link(j))
					unlinked_endpoints.push_back(TrackIter(t, j));
		}

	if(unlinked_endpoints.empty())
	{
		set_status("No free endpoints");
		set_done(false);
	}

	list<TrackType *> track_types = designer.get_catalogue().get_list<TrackType>();
	for(list<TrackType *>::iterator i=track_types.begin(); i!=track_types.end(); ++i)
	{
		const vector<TrackPart> &parts = (*i)->get_parts();
		if(parts.size()!=1)
			continue;
		if(parts.front().is_curved() || parts.front().is_dead_end())
			continue;

		types_by_length[parts.front().get_length()] = *i;
		max_preference = max(max_preference, (*i)->get_autofit_preference());
	}
}

ExtendTool::~ExtendTool()
{
	if(!accepted)
	{
		for(vector<R2C2::Track *>::iterator i=extend_tracks.begin(); i!=extend_tracks.end(); ++i)
			delete *i;
	}
}

void ExtendTool::connect()
{
	float limit;

	Track *start_track = 0;
	Track *end_track = 0;
	Snap start_sn;
	bool ok = false;
	float gap = 0;
	for(vector<TrackIter>::const_iterator i=unlinked_endpoints.begin(); i!=unlinked_endpoints.end(); ++i)
	{
		start_sn = (*i)->get_snap_node(i->entry());
		limit = (*i)->get_type().get_gauge()/10;
		
		for(vector<TrackIter>::const_iterator j=i; ++j!=unlinked_endpoints.end(); )
		{
			Snap end_sn = (*j)->get_snap_node(j->entry());

			float dz = end_sn.position.z-start_sn.position.z;
			if(abs(dz)>0.02)
				continue;

			Angle adiff = wrap_balanced(start_sn.rotation+Angle::half_turn()-end_sn.rotation);
			if(abs(adiff).radians()>0.01)
				continue;

			Vector delta = rotated_vector(end_sn.position-start_sn.position, -start_sn.rotation);
			if(abs(delta.y)>limit)
				continue;

			gap = delta.x;
			if(gap<0)
				continue;

			start_track = i->track();
			end_track = j->track();
			ok = true;
		}

		if(ok)
			break;
	}

	if(!ok)
	{
		set_status("No aligned endpoints found");
		set_done(false);
		return;
	}

	extend_tracks = create_straight(start_sn.position, start_sn.rotation, gap, limit);

	if(extend_tracks.empty())
	{
		set_status("No connection possible");
		set_done(false);
		return;
	}

	extend_tracks.front()->link_to(*start_track);
	extend_tracks.back()->link_to(*end_track);

	set_done(true);
}

void ExtendTool::pointer_motion()
{
	Vector pos;
	Angle dir;
	float length = 0;

	for(vector<TrackIter>::const_iterator i=unlinked_endpoints.begin(); i!=unlinked_endpoints.end(); ++i)
	{
		Snap sn = (*i)->get_snap_node(i->entry());
		Vector delta = rotated_vector(ground_pointer-sn.position, -sn.rotation);

		if(delta.x<length)
			continue;

		pos = sn.position;
		dir = sn.rotation;
		length = delta.x;
	}

	if(length)
	{
		vector<Track *> trks = create_straight(pos, dir, length, max(length/500, 0.001f));

		if(!trks.empty())
		{
			for(vector<Track *>::iterator i=extend_tracks.begin(); i!=extend_tracks.end(); ++i)
				delete *i;
			extend_tracks = trks;

			map<const TrackType *, unsigned> counts;
			length = 0;
			for(vector<Track *>::iterator i=extend_tracks.begin(); i!=extend_tracks.end(); ++i)
			{
				length += (*i)->get_type().get_total_length();
				++counts[&(*i)->get_type()];
			}

			string detail;
			for(map<const TrackType *, unsigned>::const_iterator i=counts.begin(); i!=counts.end(); ++i)
			{
				if(!detail.empty())
					detail += ", ";
				detail += format("%dx %s", i->second, i->first->get_article_number());
			}

			signal_status.emit(format("Extend: %.0fmm (%s)", length*1000, detail));
		}
	}
}

void ExtendTool::finish()
{
	if(!extend_tracks.empty())
	{
		for(vector<TrackIter>::const_iterator i=unlinked_endpoints.begin(); i!=unlinked_endpoints.end(); ++i)
			if(extend_tracks.front()->link_to(**i))
				break;
	}
}

vector<Track *> ExtendTool::create_straight(const Vector &start, const Angle &dir, float length, float limit)
{
	vector<float> lengths;
	unsigned preference = max_preference;
	float removed = 0;
	while(length>limit)
	{
		bool found = false;
		for(map<float, const TrackType *>::iterator i=types_by_length.end(); i!=types_by_length.begin(); )
		{
			--i;
			if(i->second->get_autofit_preference()<preference)
				continue;
			if((!removed || i->first<removed) && i->first<length+limit)
			{
				unsigned n = static_cast<unsigned>((length+limit)/i->first);
				lengths.insert(lengths.end(), n, i->first);
				length -= n*i->first;
				found = true;
				break;
			}
		}

		if(found)
			continue;

		if(lengths.empty())
		{
			if(preference>0)
			{
				--preference;
				removed = 0;
				continue;
			}
			break;
		}

		length += lengths.back();
		removed = lengths.back();
		lengths.pop_back();
	}

	vector<Track *> trks;

	if(!lengths.empty())
	{
		Vector pos = start;
		Transform trans = Transform::rotation(dir, Vector(0, 0, 1));
		for(vector<float>::iterator i=lengths.begin(); i!=lengths.end(); ++i)
		{
			Track *track = new Track(designer.get_layout(), *get_item(types_by_length, *i));
			track->set_position(pos);
			track->set_rotation(dir);

			if(!trks.empty())
				track->link_to(*trks.back());
			trks.push_back(track);

			pos += trans.transform(Vector(*i, 0, 0));
		}
	}

	return trks;
}

void ExtendTool::update_selection(Selection &sel) const
{
	if(accepted)
		sel.replace(extend_tracks.begin(), extend_tracks.end());
}
