#include <list>
#include "track.h"
#include "trackchain.h"
#include "trainroutemetric.h"

using namespace std;

namespace R2C2 {

TrainRouteMetric::TrainRouteMetric(const TrackChain &tc, TrackChain::Direction dir)
{
	vector<TrackIter> ends;
	if(dir==TrackChain::UNSPECIFIED)
	{
		for(unsigned i=0; i<2; ++i)
			if(TrackIter end = tc.get_end(i))
				ends.push_back(end);
	}
	else if(TrackIter end = tc.get_end(dir))
		ends.push_back(end);

	/* Initialize goals for the ends of the target chain.  We travel away from
	the goals in the search phase so the iters appear to point the wrong way. */
	for(vector<TrackIter>::const_iterator i=ends.begin(); i!=ends.end(); ++i)
	{
		const TrackType::Endpoint &ep = i->endpoint();
		unsigned nls = (*i)->get_n_link_slots();
		for(unsigned j=0; j<nls; ++j)
		{
			TrackIter iter(i->track(), j);
			if(ep.has_common_paths(iter.endpoint()))
				goals.push_back(iter);
		}
	}

	list<TrackIter> queue;
	for(vector<Goal>::iterator i=goals.begin(); i!=goals.end(); ++i)
	{
		tracks[Key(i->track.track(), i->track.entry())] = Data(0, &*i);
		queue.push_back(i->track);
	}

	/* Use Dijkstra's algorithm to find the shortest distance from the goal to
	every reachable track in the layout.  Entry points here become exit points
	when looking up distances to the goal. */
	while(!queue.empty())
	{
		TrackIter track = queue.front();
		queue.pop_front();
		const Data &data = tracks[Key(track.track(), track.entry())];

		const TrackType::Endpoint &ep = track.endpoint();

		float multiplier = get_travel_multiplier(*track, track.entry());
		for(unsigned i=0; ep.paths>>i; ++i)
			if(ep.has_path(i))
			{
				TrackIter next = track.next(i);
				if(!next)
					continue;

				Data &target = tracks[Key(next.track(), next.entry())];
				float dist = data.distance+track->get_type().get_path_length(i)*multiplier;
				if(target.distance<0 || target.distance>dist)
				{
					target = Data(dist, data.goal);
					queue.push_back(next);
				}
			}
	}
}

void TrainRouteMetric::chain_to(const TrainRouteMetric &metric)
{
	for(vector<Goal>::iterator i=goals.begin(); i!=goals.end(); ++i)
		i->base_distance = metric.get_distance_from(*i->track.track(), i->track.entry());
}

float TrainRouteMetric::get_distance_from(const Track &track) const
{
	map<Key, Data>::const_iterator i = tracks.lower_bound(Key(&track, 0));
	map<Key, Data>::const_iterator j = tracks.upper_bound(Key(&track, 255));

	float result = -1;
	for(; i!=j; ++i)
	{
		float d = i->second.distance+i->second.goal->base_distance;
		if(result<0 || d<result)
			result = d;
	}

	return result;
}

float TrainRouteMetric::get_distance_from(const Track &track, unsigned exit) const
{
	map<Key, Data>::const_iterator i = tracks.find(Key(&track, exit));
	if(i==tracks.end())
		return -1;

	return i->second.distance+i->second.goal->base_distance;
}

float TrainRouteMetric::get_travel_multiplier(const Track &track, unsigned exit) const
{
	int pref = track.get_preferred_exit();
	if(pref>=0 && exit!=static_cast<unsigned>(pref))
		return 5;

	return 1;
}


TrainRouteMetric::Goal::Goal():
	base_distance(0)
{ }

TrainRouteMetric::Goal::Goal(const TrackIter &t):
	track(t),
	base_distance(0)
{ }


TrainRouteMetric::Data::Data():
	distance(-1),
	goal(0)
{ }

TrainRouteMetric::Data::Data(float d, const Goal *g):
	distance(d),
	goal(g)
{ }

} // namespace R2C2
